//
// Tuner functionality
// (c) 4.12.2024 Ingo D. Rullhusen
//


#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/power.h>
#include <avr/eeprom.h>
#include <util/delay.h>


#include "usb_serial.h"
#include "bitbang_serial.h"

#include "tuner.h"
#include "diskey.h"


freq10k_fm fm_lower =  8750;
freq10k_fm fm_upper = 10800;
freq1k_am am_lower =  531;
freq1k_am am_upper = 1602;
freq_dab dab_lower = { 0, 0 };
freq_dab dab_upper = { 37, 63 };

freq10k_fm FMstation[8];
freq1k_am AMstation[8];
freq_dab DABstation[8];

uint8_t last_band, last_fm_pLED, last_am_pLED, last_dab_pLED;
freq10k_fm last_fm_freq;
freq1k_am last_am_freq;
freq_dab last_dab_freq;


void setup_to_eeprom(void)
{
   void *e = NULL;
   int s;
   s = sizeof(freq10k_fm)*8;
   eeprom_write_block(FMstation,e,s);
   e += s;
   s = sizeof(freq1k_am)*8;
   eeprom_write_block(AMstation,e,s);
   e += s;
   s = sizeof(freq_dab)*8;
   eeprom_write_block(DABstation,e,s);
   e += s;

}

void setup_from_eeprom(void)
{
   void *e = NULL;
   int s;
   s = sizeof(freq10k_fm)*8;
   eeprom_read_block(FMstation,e,s);
   e += s;
   s = sizeof(freq1k_am)*8;
   eeprom_read_block(AMstation,e,s);
   e += s;
   s = sizeof(freq_dab)*8;
   eeprom_read_block(DABstation,e,s);
   e += s;

}


void tuner_init(void)
{

// input B7:5 (NC 4:0)
   DDRB = 0x00;
//   PORTB = 0x1f; // no pull-up
   PORTB = 0xff; // pull-up

// output C2:0
   DDRC |= 0x07;
   PORTC &= ~(0x07);

// input E7:6, output E1:0
   DDRE &= ~(0xc0);
   DDRE |= 0x03;
   PORTE &= ~(0xc3);

// pull-up unused

   DDRE &= ~(0x3c);
   PORTE |= 0x38; // E2 external pull-down

// power interrupt PCINT7
   PCICR |= 0x01;
   PCMSK0 |= 0x80;

   setup_from_eeprom();

   last_band = BAND_FM;
   last_fm_freq = 0;
   last_am_freq = 0;

}


uint8_t LOCKDET(void)
{ return( PINB & 0x20 ); }

uint8_t STATDET(void)
{ return( PINB & 0x40 ); }

uint8_t PWRDET(void)   
{ return( PINB & 0x80 ); }

uint8_t CHNTUNED(void)
{ return( PINE & 0x40 ); }

uint8_t STNTUNED(void)
{ return( PINE & 0x80 ); }


void set_band( uint8_t b )
{
   if ( (PORTC & 0x07) == b ) return;

   if ( b != BAND_DAB )
   {
      uart_bitbang_send_byte( 0xff /*terminate*/ );
      _delay_ms(100);
   }
   else
   {
      diskey_str2seg( PSTR(" dAb "), 0,0);
      diskey_pLED( 0xf );
   }

   PORTC = (PORTC & 0xf8) | (0x07 & b);

   if ( b == BAND_DAB )
   {
      for ( int i=0; !CHNTUNED() && i<5000; ++i ) { _delay_ms(1); }
   }

}

void set_mute( uint8_t f )
{
   if ( f ) { PORTE |= 0x02; }
   else     { PORTE &= 0xfd; }
}

void set_power( uint8_t f )
{
   if ( f ) { PORTE |= 0x01; }
   else     { PORTE &= 0xfe; }
}


ISR(PCINT0_vect)
{
   if ( PINB & 0x80 ) // main power on
   {
      CPU_PRESCALE(2);
      power_timer2_enable();
#ifdef USB_CDC_ENABLED
      power_usb_enable();
      usb_serial_flush_input();
#endif
      MODE = 0xff;
   }
   else               // main power off
   {
      tuner_pwr_off();
      power_usb_disable();
      power_timer2_disable();
      PORTF = 0xff;
      PORTA |= 0x3f;
      MODE = M_BATPWR;
      CPU_PRESCALE(8);
   }
}


void tuner_pwr_on(void)
{
   set_mute(1);
   set_power(1);
   set_band(last_band);
   if ( last_band == BAND_MW )
   {
      tune_to_am();
      diskey_pLED( last_am_pLED );
      diskey_fm2seg( last_am_freq );
   }
   if ( last_band == BAND_FM )
   {
      tune_to_fm();
      diskey_pLED( last_fm_pLED );
      diskey_fm2seg( last_fm_freq );
   }
   if ( last_band == BAND_DAB )
   {
      tune_to_dab( 1 );
      diskey_pLED( last_dab_pLED );
      diskey_dab2seg( last_dab_freq );
   }
   set_mute(0);
}

void tuner_pwr_off(void)
{
   set_mute(1);
   diskey_pLED( 0xf );
   set_band(BAND_OFF);
   set_power(0);
}

uint8_t tuner_pwr_state(void)
{
 return( PORTE & 0x01 );
}


void tune_to_fm()
{
   if ( last_fm_freq < fm_lower ) { last_fm_freq = fm_upper; }
   if ( last_fm_freq > fm_upper ) { last_fm_freq = fm_lower; }

   freq10k_fm fpll = last_fm_freq + 1070;

   tuneFreqR[0] = 1;
   tuneFreqR[1] = 0;

   if ( fpll%10 > 4 ) { tuneFreqR[6] = 2; }
   else               { tuneFreqR[6] = 4; }
   fpll /= 10;

   for ( int i=5; i>1; --i )
   {
      tuneFreqR[i] = fpll%10;
      fpll /= 10;
   }

   tuneFreqR[7] = 7;

   diskey_set_pll = PLL_SET;
}

void tune_to_am()
{
   if ( last_am_freq < am_lower ) { last_am_freq = am_upper; }
   if ( last_am_freq > am_upper ) { last_am_freq = am_lower; }

   freq1k_am fpll = (last_am_freq + 450) / 9;

   tuneFreqR[0] = 2;
   tuneFreqR[1] = 0;

   for ( int i=4; i>1; --i )
   {
      tuneFreqR[i] = fpll%10;
      fpll /= 10;
   }

   tuneFreqR[5] = 0;
   tuneFreqR[6] = 0;

   tuneFreqR[7] = 7;

   diskey_set_pll = PLL_SET;

}

void tune_to_dab( _Bool chn_changed )
{

   if ( last_dab_freq.subchannel < dab_lower.subchannel )
   {
      chn_changed = 1;
      --last_dab_freq.channel;
      last_dab_freq.subchannel = dab_upper.subchannel;
   }
   if ( last_dab_freq.subchannel > dab_upper.subchannel )
   {
      chn_changed = 1;
      ++last_dab_freq.channel;
      last_dab_freq.subchannel = dab_lower.subchannel;
   }

   if ( last_dab_freq.channel < dab_lower.channel )
   {
      chn_changed = 1;
      last_dab_freq.channel = dab_upper.channel;
   }
   if ( last_dab_freq.channel > dab_upper.channel )
   {
      chn_changed = 1;
      last_dab_freq.channel = dab_lower.channel;
   }

   if ( chn_changed )
   {
      diskey_str2seg( PSTR("-----"), 0,1);
      diskey_pLED( 0xf );
      uart_bitbang_send_byte( 0x00 | last_dab_freq.channel );
      for ( int i=0;  CHNTUNED() && i<5000; ++i ) { _delay_ms(1); }
      for ( int i=0; !CHNTUNED() && i<5000; ++i ) { _delay_ms(1); }
   }

   uart_bitbang_send_byte( 0x80 | last_dab_freq.subchannel );

}

void tuner_wait(void)
{
   uint8_t k_t = KEY;
   if ( k_t > K_0_MW && k_t < K_9_FM )
   {
      if      ( last_band == BAND_DAB )
      {
         last_dab_freq = DABstation[ k_t - K_1 ];
         last_dab_pLED = k_t - K_1;
      }
      else if ( last_band == BAND_MW  )
      {
         last_am_freq = AMstation[ k_t - K_1 ];
         last_am_pLED = k_t - K_1;
      }
      else
      {
         last_band = BAND_FM;
         last_fm_freq = FMstation[ k_t - K_1 ];
         last_fm_pLED = k_t - K_1;
      }
      tuner_pwr_on();
   }
   else
   {
      diskey_time2seg( (bcd_time*)&CLOCK );
      diskey_pLED(0xff);
      set_band(BAND_OFF);
   }
}

void change_dab_stn( int8_t step )
{
   uint8_t mode_t = MODE;
   uint8_t key_t;

   set_mute(1);

   if ( step > 0 ) { key_t = K_RIGHT; }
   else            { key_t = K_LEFT;  }
   last_dab_freq.subchannel += step;
   tune_to_dab( 0 );
   diskey_pLED( last_dab_pLED = 0xf );
   diskey_dab2seg( last_dab_freq );
   _delay_ms(200);

   if ( KEY == key_t || KEY == K_NONE ) { _delay_ms(1000); }

   while ( ( ( KEY == key_t ) ||
             ( AUTOSCAN && !STNTUNED() && (KEY == key_t || KEY == K_NONE ) )
           ) && mode_t == MODE )
   {
      if ( AUTOSCAN )
      {
         if ( CHNTUNED() )
         {
            last_dab_freq.subchannel += step;
         }
         else
         {
            if      ( step < 0 )
            { last_dab_freq.subchannel = dab_lower.subchannel-1; }
            else if ( step > 0 )
            { last_dab_freq.subchannel = dab_upper.subchannel+1; }
         }
      }
      else
      {
         last_dab_freq.subchannel += step;
      }
      tune_to_dab( 0 );
      diskey_dab2seg( last_dab_freq );
      _delay_ms(200);
   }

   set_mute(0);

}

void change_am_freq( freq1k_am step )
{
   uint8_t mode_t = MODE;
   uint8_t key_t;

   set_mute(1);

   if ( step > 0 ) { key_t = K_RIGHT; }
   else            { key_t = K_LEFT;  }
   last_am_freq += step;
   tune_to_am();
   diskey_pLED( last_am_pLED = 0xf );
   diskey_am2seg( last_am_freq );
   for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
   _delay_ms(50);

   if ( KEY == key_t || KEY == K_NONE ) { _delay_ms(1000); }

   while ( ( ( KEY == key_t ) ||
             ( AUTOSCAN && !STATDET() && (KEY == key_t || KEY == K_NONE ) )
           ) && mode_t == MODE )
   {
      last_am_freq += step;
      tune_to_am();
      diskey_am2seg( last_am_freq );
      for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
      _delay_ms(50);
   }

   set_mute(0);

}

void change_fm_freq( freq10k_fm step )
{
   uint8_t mode_t = MODE;
   uint8_t key_t;

   set_mute(1);

   if ( step > 0 ) { key_t = K_RIGHT; }
   else            { key_t = K_LEFT;  }
   last_fm_freq += step;
   tune_to_fm();
   diskey_pLED( last_fm_pLED = 0xf );
   diskey_fm2seg( last_fm_freq );
   for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
   _delay_ms(50);

   if ( KEY == key_t || KEY == K_NONE ) { _delay_ms(1000); }

   while ( ( ( KEY == key_t ) ||
             ( AUTOSCAN && !STATDET() && (KEY == key_t || KEY == K_NONE ) )
           ) && mode_t == MODE )
   {
      last_fm_freq += step;
      tune_to_fm();
      diskey_fm2seg( last_fm_freq );
      for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
      _delay_ms(50);
   }

   set_mute(0);

}

void tuner_on(void)
{
   static uint8_t last_k = K_NONE;
   uint8_t k_t = KEY;

   if ( k_t == last_k )
   {
      if ( last_band == BAND_MW )
      {
         diskey_am2seg( last_am_freq );
         diskey_pLED( last_am_pLED );
      }
      if ( last_band == BAND_FM )
      {
         diskey_fm2seg( last_fm_freq );
         diskey_pLED( last_fm_pLED );
      }
      if ( last_band == BAND_DAB )
      {
         diskey_dab2seg( last_dab_freq );
         diskey_pLED( last_dab_pLED );
      }

      return;
   }
   last_k = k_t;

   if ( k_t == K_0_MW && (last_band!=BAND_MW) )
   {
      set_mute(1);
      set_band( (last_band=BAND_MW) );
      tune_to_am();
      diskey_pLED( last_am_pLED );
      diskey_am2seg( last_am_freq );
      for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
      _delay_ms(50);
      set_mute(0);
      return;
   }

   if ( k_t == K_9_FM && (last_band!=BAND_FM) )
   {
      set_mute(1);
      set_band( (last_band=BAND_FM) );
      tune_to_fm();
      diskey_pLED( last_fm_pLED );
      diskey_fm2seg( last_fm_freq );
      for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
      _delay_ms(50);
      set_mute(0);
      return;
   }

   if ( ( k_t == K_0_MW || k_t == K_9_FM ) && (last_band!=BAND_DAB) )
   {
      set_mute(1);
      set_band( (last_band=BAND_DAB) );
      tune_to_dab( 1 );
      diskey_pLED( last_dab_pLED );
      diskey_dab2seg( last_dab_freq );
      _delay_ms(100);
      set_mute(0);
      return;
   }

   if ( k_t > K_0_MW && k_t <= K_9_FM )
   {
      set_mute(1);
      if ( last_band == BAND_DAB )
      {
         last_dab_freq = DABstation[ k_t - K_1 ];
         tune_to_dab( 1 );
         diskey_pLED( (last_dab_pLED=k_t-K_1) );
         diskey_dab2seg( last_dab_freq );
         _delay_ms(50);
      }
      else if ( last_band == BAND_MW )
      {
         last_am_freq = AMstation[ k_t - K_1 ];
         tune_to_am();
         diskey_pLED( (last_am_pLED=k_t-K_1) );
         diskey_am2seg( last_am_freq );
         for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
      }
      else
      {
         last_fm_freq = FMstation[ k_t - K_1 ];
         tune_to_fm();
         diskey_pLED( (last_fm_pLED=k_t-K_1) );
         diskey_fm2seg( last_fm_freq );
         for ( int i=0; !LOCKDET() && i<1000; ++i ) { _delay_ms(1); }
      }
      _delay_ms(50);
      set_mute(0);
      return;
   }

   if ( k_t == K_STORE )
   {
      diskey_sLED(1);
      for ( int i=0; i<4000; ++i )
      {
         k_t = KEY;

         if ( k_t > K_0_MW && k_t <= K_9_FM )
         {
            if ( last_band == BAND_MW )
            {
               AMstation[ k_t - K_1 ] = last_am_freq;
               diskey_pLED( (last_am_pLED=k_t-K_1) );
               setup_to_eeprom();
               break;
            }
            if ( last_band == BAND_FM )
            {
               FMstation[ k_t - K_1 ] = last_fm_freq;
               diskey_pLED( (last_fm_pLED=k_t-K_1) );
               setup_to_eeprom();
               break;
            }
            if ( last_band == BAND_DAB )
            {
               DABstation[ k_t - K_1 ] = last_dab_freq;
               diskey_pLED( (last_dab_pLED=k_t-K_1) );
               setup_to_eeprom();
               break;
            }
         }
         _delay_ms(1);
      }
      diskey_sLED(0);

      return;
   }

   if ( k_t == K_LEFT )
   {
      if      ( last_band == BAND_DAB ) { change_dab_stn( -1 ); }
      else if ( last_band == BAND_MW  ) { change_am_freq( -9 ); }
      else                              { change_fm_freq( -5 ); }
      return;
   }

   if ( k_t == K_RIGHT )
   {
      if      ( last_band == BAND_DAB ) { change_dab_stn( +1 ); }
      else if ( last_band == BAND_MW  ) { change_am_freq( +9 ); }
      else                              { change_fm_freq( +5 ); }
      return;
   }

}


void show_message()
{
   diskey_str2seg( PSTR(" inGo"), 0,0);
   for ( int i=0; MODE==M_MSG && i<150; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("     "), 0,0);
   for ( int i=0; MODE==M_MSG && i<50; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("  d  "), 0,1);
   for ( int i=0; MODE==M_MSG && i<100; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("     "), 0,0);
   for ( int i=0; MODE==M_MSG && i<50; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   const uint8_t st = 25;
   diskey_str2seg( PSTR("    r"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("   ru"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("  ruL"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR(" ruLL"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("ruLLh"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("uLLhu"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("LLhus"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("LhusE"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("husEn"), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("usEn "), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("sEn  "), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("En   "), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("n    "), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

   diskey_str2seg( PSTR("     "), 0,0);
   for ( int i=0; MODE==M_MSG && i<st; ++i ) { _delay_ms(10); }
   if ( MODE != M_MSG ) return;

}

